package reconciler

import (
	"context"
	"sort"

	"github.com/hashicorp/go-version"
	"k8s.libre.sh/application"
	interfaces "k8s.libre.sh/interfaces"
	"k8s.libre.sh/status"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
)

type InstanceReconciler struct {
	*ReconcilerBase
	Instance  application.Instance
	SyncOrder map[int]string
	//	ctx            context.Context
}

func NewInstanceReconciler(r *ReconcilerBase, i application.Instance, order map[int]string) *InstanceReconciler {
	return &InstanceReconciler{
		ReconcilerBase: r,
		Instance:       i,
		SyncOrder:      order,
	}
}

func (r *InstanceReconciler) Reconcile(request reconcile.Request) (reconcile.Result, error) {

	appStatus := r.Instance.GetApplicationStatus()
	oldStatus := appStatus.DeepCopy()

	err := application.Init(r.Instance, r.ReconcilerBase)
	if err != nil {
		return reconcile.Result{}, err
	}

	// Sync Settings
	for name, s := range r.Instance.GetSettings() {

		res, err := SyncComponent(context.TODO(), r.ReconcilerBase, s, r.Instance.GetOwner())
		cptStatus := status.ComponentStatus{
			Resources: res,
		}
		if len(appStatus.Settings) == 0 {
			appStatus.Settings = make(map[string]status.ComponentStatus)
		}
		appStatus.Settings[name] = cptStatus
		if err != nil {
			cptStatus.ComponentsReady = "NotReady"
			return ctrl.Result{}, err
		}
		cptStatus.ComponentsReady = "Ready"
	}

	jobs := r.Instance.GetJobs()

	current, err := version.NewVersion(r.Instance.GetApplicationStatus().Version)
	// get nextcloud desired version in spec
	desired, err := version.NewVersion(r.Instance.GetVersion())

	if len(r.Instance.GetApplicationStatus().Version) == 0 {
		delete(jobs, "upgrade")
	} else {
		switch current.Compare(desired) {
		case 1:
			// current version is greater than desired, no downgrade
			// TODO Format error
			return ctrl.Result{}, err
		case -1:
			// current version lower than desired, upgrade job

			delete(jobs, "install")
		case 0:
			// current version and desired are the same, normal start
			delete(jobs, "install")
			delete(jobs, "upgrade")
		}
	}

	cpts := make(map[string]application.ComponentMutate)

	for name, j := range jobs {
		cpts[name] = j
	}

	// Sync Components
	for name, c := range r.Instance.GetComponents() {
		cpts[name] = c
	}

	res, err := Sync(context.TODO(), r.ReconcilerBase, cpts, r.Instance.GetOwner(), r.SyncOrder)

	appStatus.Components = res.Components
	appStatus.Conditions = res.Conditions

	if err != nil {
		return ctrl.Result{}, err
	}

	// Clean up resources and update status
	err = CleanUpResources(oldStatus.GetResources(), appStatus.GetResources(), r.ReconcilerBase.GetClient(), r.Instance.GetNamespace())
	if err != nil {
		return ctrl.Result{}, err
	}
	appStatus.Version = desired.String()
	owner := r.Instance.GetOwner()
	owner.SetApplicationStatus(appStatus)

	if err := r.ReconcilerBase.GetClient().Status().Update(context.TODO(), owner); err != nil {
		return ctrl.Result{}, err
	}

	return ctrl.Result{}, nil
}

func Sync(ctx context.Context, r interfaces.Reconcile, cpts map[string]application.ComponentMutate, owner interfaces.Object, order map[int]string) (appStatus status.ApplicationStatus, err error) {

	keys := []int{}
	appStatus.Components = make(map[string]status.ComponentStatus)

	for k := range order {
		keys = append(keys, k)
	}

	sort.Ints(keys)

	for _, k := range keys {
		if cpts[order[k]] != nil {
			resources, err := SyncComponent(ctx, r, cpts[order[k]], owner)
			appStatus.Components[order[k]] = status.ComponentStatus{
				Resources: resources,
			}
			if err != nil {
				// manage error
				appStatus.Conditions = append(appStatus.Conditions, status.ReconcileError(err))
				return appStatus, err
			}
		}
	}

	appStatus.Conditions = append(appStatus.Conditions, status.ReconcileSuccess())

	return appStatus, nil
}
