/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package status

import (
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.libre.sh/interfaces"
)

// Phase is the current status of a App as a whole.
type Phase string

const (
	PhaseNone       Phase = ""
	PhasePlanning   Phase = "Planning"
	PhaseRunning    Phase = "Running"
	PhaseCreating   Phase = "Creating"
	PhaseInstalling Phase = "Installing"
	PhaseUpgrading  Phase = "Upgrading"
	PhaseComplete   Phase = "Complete"
	PhaseFailed     Phase = "Failed"
)

// ConditionType encodes information on the condition
// +kubebuilder:object:generate=true
type ConditionType string

const (
	ConditionNone       ConditionType = ""
	ConditionPlanning   ConditionType = "Planning"
	ConditionRunning    ConditionType = "Running"
	ConditionCreating   ConditionType = "Creating"
	ConditionInstalling ConditionType = "Installing"
	ConditionUpgrading  ConditionType = "Upgrading"
	ConditionComplete   ConditionType = "Complete"
	ConditionFailed     ConditionType = "Failed"
)

// A ConditionReason represents the reason a resource is in a condition.
type ConditionReason string

// Reasons a resource is or is not ready.
const (
	ReasonAvailable   ConditionReason = "Resource is available for use"
	ReasonUnavailable ConditionReason = "Resource is not available for use"
	ReasonCreating    ConditionReason = "Resource is being created"
	ReasonDeleting    ConditionReason = "Resource is being deleted"
)

// Reasons a resource is or is not synced.
const (
	ReasonReconcileSuccess ConditionReason = "Successfully reconciled resource"
	ReasonReconcileError   ConditionReason = "Encountered an error during resource reconciliation"
)

// Condition types.
const (
	// TypeReady resources are believed to be ready to handle work.
	TypeReady ConditionType = "Ready"

	// TypeSynced resources are believed to be in sync with the
	// Kubernetes resources that manage their lifecycle.
	TypeSynced ConditionType = "Synced"
)

// Creating returns a condition that indicates the resource is currently
// being created.
func Creating() Condition {
	return Condition{
		Type:               TypeReady,
		Status:             corev1.ConditionFalse,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonCreating,
	}
}

// Deleting returns a condition that indicates the resource is currently
// being deleted.
func Deleting() Condition {
	return Condition{
		Type:               TypeReady,
		Status:             corev1.ConditionFalse,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonDeleting,
	}
}

// Available returns a condition that indicates the resource is
// currently observed to be available for use.
func Available() Condition {
	return Condition{
		Type:               TypeReady,
		Status:             corev1.ConditionTrue,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonAvailable,
	}
}

// Unavailable returns a condition that indicates the resource is not
// currently available for use. Unavailable should be set only when Crossplane
// expects the resource to be available but knows it is not, for example
// because its API reports it is unhealthy.
func Unavailable() Condition {
	return Condition{
		Type:               TypeReady,
		Status:             corev1.ConditionFalse,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonUnavailable,
	}
}

// ReconcileSuccess returns a condition indicating that Crossplane successfully
// completed the most recent reconciliation of the resource.
func ReconcileSuccess() Condition {
	return Condition{
		Type:               TypeSynced,
		Status:             corev1.ConditionTrue,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonReconcileSuccess,
	}
}

// ReconcileError returns a condition indicating that Crossplane encountered an
// error while reconciling the resource. This could mean Crossplane was
// unable to update the resource to reflect its desired state, or that
// Crossplane was unable to determine the current actual state of the resource.
func ReconcileError(err error) Condition {
	return Condition{
		Type:               TypeSynced,
		Status:             corev1.ConditionFalse,
		LastTransitionTime: metav1.Now(),
		Reason:             ReasonReconcileError,
		Message:            err.Error(),
	}
}

// ApplicationStatus defines controller's the observed state of Application
// +kubebuilder:object:generate=true
type ApplicationStatus struct {
	Version string `json:"version,omitempty"`
	// Phase   Phase  `json:"phase,omitempty"`
	// ObservedGeneration is the most recent generation observed. It corresponds to the
	// Object's generation, which is updated on mutation by the API Server.
	ObservedGeneration int64 `json:"observedGeneration,omitempty" protobuf:"varint,1,opt,name=observedGeneration"`
	// Conditions represents the latest state of the object
	Conditions []Condition `json:"conditions,omitempty" patchStrategy:"merge" patchMergeKey:"type" protobuf:"bytes,10,rep,name=conditions"`
	// Components is a list of object ComponentStatus
	Components map[string]ComponentStatus `json:"components,omitempty"`
	Settings   map[string]ComponentStatus `json:"settings,omitempty"`
	// ComponentsReady: status of the components in the format ready/total
	ComponentsReady string `json:"componentsReady,omitempty"`
}

func (app *ApplicationStatus) GetResources() *Resources {
	resources := &Resources{}

	for _, c := range app.Components {
		resources.Objects = append(resources.Objects, c.GetObjects()...)
	}

	return resources
}

// ApplicationStatus defines controller's the observed state of Application
// +kubebuilder:object:generate=true
type ComponentStatus struct {
	Name string `json:"name,omitempty"`
	// Resources embeds a list of object statuses
	Resources `json:",inline,omitempty"`
	// ComponentsReady: status of the components in the format ready/total
	ComponentsReady string `json:"componentsReady,omitempty"`
}

// Condition describes the state of an object at a certain point.
// +kubebuilder:object:generate=true
type Condition struct {
	// Type of condition.
	Type ConditionType `json:"type" protobuf:"bytes,1,opt,name=type,casttype=StatefulSetConditionType"`
	// Status of the condition, one of True, False, Unknown.
	Status corev1.ConditionStatus `json:"status" protobuf:"bytes,2,opt,name=status,casttype=k8s.io/api/core/v1.ConditionStatus"`
	// The reason for the condition's last transition.
	Reason ConditionReason `json:"reason,omitempty" protobuf:"bytes,4,opt,name=reason"`
	// A human readable message indicating details about the transition.
	Message string `json:"message,omitempty" protobuf:"bytes,5,opt,name=message"`
	// Last time the condition was probed
	//	LastUpdateTime metav1.Time `json:"lastUpdateTime,omitempty" protobuf:"bytes,3,opt,name=lastProbeTime"`
	// Last time the condition transitioned from one status to another.
	LastTransitionTime metav1.Time `json:"lastTransitionTime,omitempty" protobuf:"bytes,3,opt,name=lastTransitionTime"`
}

// ComponentList is a generic status holder for the top level resource
// +kubebuilder:object:generate=true
type Resources struct {
	// Object status array for all matching objects
	Objects []ObjectStatus `json:"objects,omitempty"`
}

func Difference(original, desired *Resources) *Resources {
	// diff := original.DeepCopy()

	diff := &Resources{}
	diffByKey := make(map[string]ObjectStatus, len(original.Objects))
	desiredByKey := make(map[string]ObjectStatus, len(desired.Objects))

	for _, obj := range original.Objects {
		diffByKey[obj.Link] = obj
	}

	for _, obj := range desired.Objects {
		desiredByKey[obj.Link] = obj
	}

	for item := range desiredByKey {
		delete(diffByKey, item)
	}

	for _, obj := range diffByKey {
		diff.Objects = append(diff.Objects, obj)
	}

	return diff
}

func (res *Resources) GetObjects() []ObjectStatus {
	return res.Objects
}

// ObjectStatus is a generic status holder for objects
// +kubebuilder:object:generate=true
type ObjectStatus struct {
	// Link to object
	Link string `json:"link,omitempty"`
	// Name of object
	Name string `json:"name,omitempty"`
	// Kind of object
	Version string `json:"version,omitempty"`
	// Kind of object
	Kind string `json:"kind,omitempty"`
	// Object group
	Group string `json:"group,omitempty"`
	// Status. Values: InProgress, Ready, Unknown
	Status string `json:"status,omitempty"`
}

func (ost *ObjectStatus) GVK() schema.GroupVersionKind {
	return schema.GroupVersionKind{
		Group:   ost.Group,
		Version: ost.Version,
		Kind:    ost.Kind,
	}
}

func ObjectToStatus(obj interfaces.Object) ObjectStatus {
	return ObjectStatus{
		Link:    obj.GetSelfLink(),
		Name:    obj.GetName(),
		Group:   obj.GetObjectKind().GroupVersionKind().Group,
		Version: obj.GetObjectKind().GroupVersionKind().Version,
		Kind:    obj.GetObjectKind().GroupVersionKind().Kind,
	}
}

// Object interface must be supported by all types that want to sync an object.
// The object interface provides a mutate function and a runtime.Object that can be used in controller-runtime CreateOrUpdate
type ObjectWithStatus interface {
	interfaces.Object
	GetApplicationStatus() ApplicationStatus
	SetApplicationStatus(conditions ApplicationStatus)
}
